#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <ctype.h>
#include "tokenizer.h"

#define LOOKING_FOR_START_OF_TOKEN	1
#define LOOKING_FOR_END_OF_TOKEN	2

// prototypes of functions we do not necessarily want to put in tokenizer.h
// since we don't expect other programmers to use them directly 
// they are only useful for us to break down our job
void 	tokens_add(Tokens* p, char* user_input, int start, int stop); 
double 	tokens_evaluate_(Tokens* p, int * last);
int 	token_verify(Token* p);
void 	token_display(Token* p); 
void 	token_deallocate(Token* p);
void 	token_identify(Token* p); 


void token_identify(Token* p){ 

	if (p == NULL){ 
		printf("ERROR: trying to identify NULL token\n"); 
		fflush(stdout); exit(EXIT_FAILURE); }
	
	int digits 	= 0; 
	int letters 	= 0; 
	int dot		= 0; 
	
	if (p->str == NULL){
		printf("ERROR: trying to identify Token w/ NULL string\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	int i; 
	int length = strlen(p->str); 
	for(i = 0 ; i < length ; i++){ 
		if (isdigit(p->str[i]))		digits++; 
		if (p->str[i] == '.')		dot++; 
		if (isalpha(p->str[i]))		letters++; 
	}
	
	p->type = UNKNOWN; 
	
	if (dot > 1)							p->type = UNKNOWN; 
	else if ( (length == 1) && ( 
			(p->str[0] == '+') || 
			(p->str[0] == '-') || 
			(p->str[0] == '/') ||
			(p->str[0] == '*') ) )				p->type = OPERATOR; 
	else if (length == letters) 					p->type = TEXT; 
	else if (length == digits)					p->type = INTEGER; 
	else if ( (letters == 0) && (dot == 1) && (digits > 0) )	p->type = FLOATING; 
	
	
	switch (p->type) {
		case INTEGER: 	sscanf(p->str, "%d", &(p->value.integer)); break;
		case FLOATING:	sscanf(p->str, "%lf", &(p->value.floating)); break;
		default: 	break;
	}
	
}

double tokens_evaluate(Tokens* p){
	// locate the last element in *p
	int last; 
	for(last = 0 ; p->tokens[last] != NULL ; last++); 
	
	last --; 
	
	if ( last < 0 ) { 
		printf("ERROR: no tokens left\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	
	return tokens_evaluate_(p, &last); 
}


double tokens_evaluate_(Tokens* p, int* last)
{
	if ( *last < 0 ) { 
		printf("ERROR: no tokens left\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}

	double val; 
	
	if (p->tokens[*last]->type == INTEGER) 	{	val = (double)((p->tokens[*last])->value.integer); 
							(*last)--; 
							return val; 
						}
	if (p->tokens[*last]->type == FLOATING)	{	val = (p->tokens[*last])->value.floating;
							(*last)--; 
							return val;
						}
	if (p->tokens[*last]->type == OPERATOR){ 
	
		// handling operand #1 
		int opindex = (*last)--; 
		double op1 = tokens_evaluate_(p, last); 
		// handling operand #2 
		double op2 = tokens_evaluate_(p, last); 
		switch (p->tokens[opindex]->str[0]){
			case '+': 	val = op1 + op2; break; 
			case '-':	val = op1 - op2; break; 
			case '/':	val = op1 / op2; break; 
			case '*':	val = op1 * op2; break; 
			default:	printf("ERROR: unknown operator %c\n", p->tokens[opindex]->str[0]); 
					fflush(stdout); 
					exit(EXIT_FAILURE); 		
		}
		return val;	
	}
	printf("ERROR: unexpected token type;"); 
	token_display(p->tokens[*last]); 
	printf("\n");
	fflush(stdout); 
	exit(EXIT_FAILURE);	
}




int tokens_verify(Tokens* p){
	int i; 
	for( i = 0 ; i < p->maxnbtokens ; i++){
		if (p->tokens[i] == NULL)		return 1; 
		if (! token_verify(p->tokens[i]))	return 0; 
	}
	return 1; 
}


int token_verify(Token* p){ 
	int val = ((p->type == INTEGER) || (p->type == FLOATING) || (p->type == OPERATOR));
	#ifdef DEBUG
	printf("Verifying token [%d]; ", val); 
	token_display(p); 
	printf("\n"); 
	#endif
	return val;
}


void tokens_identify(Tokens* p){
	int i; 
	for( i = 0 ; i < p->maxnbtokens ; i++){
		if (p->tokens[i])	token_identify(p->tokens[i]); 
		else			break;
	}
}

void tokenize(Tokens* p, char * str)
{
	#ifdef DEBUG_MODE
	printf("\tTokenizing: %s\n", str); 
	#endif
	
	if (str == NULL)	return; 
	if (p == NULL){ 
		printf("ERROR: Trying to tokenize to a null Tokens* pointer\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	
	
	
	int length; 
	length = strlen(str); 
	
	// let's start our search for tokens in the str string
	int 	start 	= -1; 
	int 	stop 	= -1; 
	int 	i 	= 0; 
	int	status	= LOOKING_FOR_START_OF_TOKEN; 
	
	do { 
	
		if ((str[i] == ' ') || (str[i] == '\t') || (str[i] == '\n') || (str[i] == '\0')){
			
			// we just detected a "separator" character
			
			if (status == LOOKING_FOR_START_OF_TOKEN){
				// we are looking for a start of token
				// this won't be here
			} else if (status == LOOKING_FOR_END_OF_TOKEN){
				// we already noticed the index where a token started
				// it just ended 
				stop = i -1; 
				
				// let's process the token we just identified
				tokens_add(p, str, start, stop);
				
				// we will now keep looking for the next token
				status = LOOKING_FOR_START_OF_TOKEN; 
			}
			
		} else {
		
			// we are on a character that is part of a token
			if (status == LOOKING_FOR_START_OF_TOKEN){
				// we are looking for a start of token
				// this is it
				start = i; 
				
				// now, let's keep looking at other characters 
				// until we encounter a "separator" 
				// that will tell us we found the end of this token
				status = LOOKING_FOR_END_OF_TOKEN; 
			} else if (status == LOOKING_FOR_END_OF_TOKEN){
				// we already noticed the index where a token started
				// this character is just part of it

			}
		
		}
		
		i++; 
	
	} while (i < length); 
	
	tokens_identify(p);
	#ifdef DEBUG_MODE
	printf("\tDONE tokenizing %s\n", str); 
	#endif
}



void token_display(Token* p){


	printf("\t(type = "); 
	switch (p->type){ 
		case OPERATOR: 	printf("operator"); break; 
		case INTEGER: 	printf("integer "); break; 
		case FLOATING: 	printf("floating"); break; 
		case TEXT: 	printf("text    "); break; 
		case NOTSET: 	printf("not set "); break; 
		default: 	printf("unknown "); break; 				
	}
	switch (p->type){
		case INTEGER:	printf(", value = %d)\t",p->value.integer); break; 
		case FLOATING:	printf(", value = %.2lf)\t", p->value.floating); break;
		default: 	printf(")\t\t"); 
	}
	printf("= [%s]", p->str);
}


void tokens_display(Tokens* p)
{
	if (p == NULL){ 
		printf("ERROR: Trying to display null pointer on Tokens\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	
	printf("\tIdentified Tokens:\n"); 
	int i; 
	for(i=0 ; i < p->maxnbtokens ; i++)
		if (p->tokens[i] != NULL){
			printf("\tTokens[%d]", i); 
			token_display(p->tokens[i]); 
			printf("\n"); 
		} else
			return; 
}



Tokens* tokens_allocate(int maxnbtokens)
{
	// let's allocate the data structure to contain all our tokens copies
		
	if (maxnbtokens <= 0){ 
		printf("ERROR: trying to allocate negative or null number of Tokens\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	
	Tokens* ptr = malloc(sizeof(Tokens)); 
	
	if (ptr == NULL){
		perror("Couldn't allocate tokens struct"); 
		exit(EXIT_FAILURE); 
	}
	
	ptr->tokens= malloc(sizeof(Token*) * maxnbtokens); 
	ptr->maxnbtokens = maxnbtokens;
	
	if (ptr->tokens == NULL){
		perror("Couldn't allocate tokens array"); 
		exit(-1); 
	}
	
	// we have now an array holding n pointers on strings
	// each time we discover a new token we will allocate a string 
	// to hold a copy of it and put its address at the next available 
	// pointer in strings_array. To make it easy to identify which pointer(s)
	// are free, we will set them all to NULL initially
	
	int n; 
	for(n=0 ; n < ptr->maxnbtokens; n++)
		ptr->tokens[n] = NULL; 
	
	return ptr; 
}

void tokens_deallocate(Tokens* p)
{
	#ifdef DEBUG_MODE
	printf("\tDeallocating tokens...\n"); fflush(stdout); 
	#endif 
	
	int i; 
	for(i=0 ; i < p->maxnbtokens ; i++)
		if (p->tokens[i] != NULL){ 	token_deallocate(p->tokens[i]); 
						p->tokens[i] = NULL;
		} else				break;
	
	free(p->tokens); 
	free(p); 
	#ifdef DEBUG_MODE
	printf("\tDONE deallocating tokens\n"); fflush(stdout);
	#endif
}

void token_deallocate(Token* p)
{
	if (p == NULL){ 
		printf("ERROR: trying to deallocate NULL Token*\n");
		fflush(stdout); 
		exit(EXIT_FAILURE);
	}
	if (p->str == NULL){
		printf("ERROR: trying to free NULL string in Token\n"); 
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	free(p->str); 
	free(p);
}


void tokens_add(Tokens* p, 
		char* user_input, 
		int start, int stop)
{
	int n = 0; 
	
	// look for the first unused pointer in strings_array 
	for ( n = 0 ; (n < p->maxnbtokens) && (p->tokens[n] != NULL) ; n++); 
	
	if (n == p->maxnbtokens){ 
		printf("ERROR: Exhausted max number of tokens (%d) in tokens_add function\n", n);
		fflush(stdout); 
		exit(EXIT_FAILURE); 
	}
	
	// let's measure the length of the token we found in user_input
	int token_size  = stop - start + 1 + 1; // for the '\0' we'll have to add

	// allocating a dynamical string to hold this token's copy
	char* token_itself = malloc(sizeof(char) * token_size); 
	
	//let's make a copy of the token we found in the user_input string
	// we copy all the characters between start and stop (both inclusive)
	strncpy(token_itself, user_input + start, token_size-1); 
	
	// then we add a '\0' to make sure our newly created strings is properly
	// formatted
	token_itself[token_size-1] = '\0'; 
	
	// we add this new string to the strings_array
	if ((p->tokens[n] = malloc(sizeof(Token))) == NULL){
		perror("unable to allocate Token");
		exit(EXIT_FAILURE); 
	}
	p->tokens[n]->str = token_itself; 
	p->tokens[n]->type = NOTSET; 
	
	#ifdef DEBUG_MODE
	printf("\tJust adding token at index %d start = %d stop = %d  [%s] ...\n", n, start, stop, p->tokens[n]); 
	#endif
	
}
